@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2022 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -21,7 +20,7 @@
 #include <fluent-bit/flb_info.h>
 #include <fluent-bit/flb_filter_plugin.h>
 #include <fluent-bit/flb_compat.h>
-#include <fluent-bit/flb_hash.h>
+#include <fluent-bit/flb_hash_table.h>
 #include <fluent-bit/flb_regex.h>
 #include <fluent-bit/flb_io.h>
 #include <fluent-bit/flb_upstream.h>
@@ -46,7 +45,6 @@
 #define FLB_KUBE_META_INIT_CONTAINER_STATUSES_KEY_LEN \
     (sizeof(FLB_KUBE_META_INIT_CONTAINER_STATUSES_KEY) - 1)
 #define FLB_KUBE_TOKEN_BUF_SIZE 8192       /* 8KB */
-#define FLB_KUBE_TOKEN_TTL 600             /* 10 minutes */
 
 static int file_to_buffer(const char *path,
                           char **out_buf, size_t *out_size)
@@ -110,7 +108,7 @@ static int get_token_with_command(const char *command,
     res = flb_calloc(1, FLB_KUBE_TOKEN_BUF_SIZE);
     if (!res) {
         flb_errno();
-        fclose(fp);
+        pclose(fp);
         return -1;
     }
     
@@ -121,7 +119,7 @@ static int get_token_with_command(const char *command,
             if (temp == NULL) {
                 flb_errno();
                 flb_free(res);
-                fclose(fp);
+                pclose(fp);
                 return -1;
             }
             res = temp;
@@ -132,11 +130,11 @@ static int get_token_with_command(const char *command,
 
     if (strlen(res) < 1) {
         flb_free(res);
-        fclose(fp);
+        pclose(fp);
         return -1;
     }
     
-    fclose(fp);
+    pclose(fp);
 
     *out_buf = res;
     *out_size = strlen(res);
@@ -162,17 +160,15 @@ static int get_http_auth_header(struct flb_kube *ctx)
         if (ret == -1) {
             flb_plg_warn(ctx->ins, "failed to run command %s", ctx->kube_token_command);
         }
-        ctx->kube_token_create = time(NULL);
-    } 
+    }
     else {
         ret = file_to_buffer(ctx->token_file, &tk, &tk_size);
         if (ret == -1) {
             flb_plg_warn(ctx->ins, "cannot open %s", FLB_KUBE_TOKEN);
         }
-        /* Token from token file will not expire */
-        /* Set the creation time to 0 to aviod refresh */
-        ctx->kube_token_create = 0;
+        flb_plg_info(ctx->ins, " token updated");
     }
+    ctx->kube_token_create = time(NULL);
 
     /* Token */
     if (ctx->token != NULL) {
@@ -211,19 +207,17 @@ static int refresh_token_if_needed(struct flb_kube *ctx)
     int expired = 0;
     int ret;
 
-    if (ctx->kube_token_command != NULL) {
-        if (ctx->kube_token_create > 0) {
-            if (time(NULL) > ctx->kube_token_create + FLB_KUBE_TOKEN_TTL) {
-                expired = FLB_TRUE;
-            }
+    if (ctx->kube_token_create > 0) {
+        if (time(NULL) > ctx->kube_token_create + ctx->kube_token_ttl) {
+            expired = FLB_TRUE;
         }
-        
-        if (expired || ctx->kube_token_create == 0) {
-            ret = get_http_auth_header(ctx);
-            if (ret == -1) {
-                flb_plg_warn(ctx->ins, "failed to set http auth header");
-                return -1;
-            }
+    }
+
+    if (expired || ctx->kube_token_create == 0) {
+        ret = get_http_auth_header(ctx);
+        if (ret == -1) {
+            flb_plg_warn(ctx->ins, "failed to set http auth header");
+            return -1;
         }
     }
 
@@ -337,7 +331,8 @@ static int get_meta_file_info(struct flb_kube *ctx, const char *namespace,
 
         if (payload_size) {
             packed = flb_pack_json(payload, payload_size,
-                                   buffer, size, root_type);
+                                   buffer, size, root_type,
+                                   NULL);
         }
 
         if (payload) {
@@ -359,7 +354,7 @@ static int get_meta_info_from_request(struct flb_kube *ctx,
                                       char* uri)
 {
     struct flb_http_client *c;
-    struct flb_upstream_conn *u_conn;
+    struct flb_connection *u_conn;
     int ret;
     size_t b_sent;
     int packed;
@@ -378,6 +373,7 @@ static int get_meta_info_from_request(struct flb_kube *ctx,
     ret = refresh_token_if_needed(ctx);
     if (ret == -1) {
         flb_plg_error(ctx->ins, "failed to refresh token");
+        flb_upstream_conn_release(u_conn);
         return -1;
     }
     
@@ -409,7 +405,7 @@ static int get_meta_info_from_request(struct flb_kube *ctx,
     }
 
     packed = flb_pack_json(c->resp.payload, c->resp.payload_size,
-                                   buffer, size, root_type);
+                                   buffer, size, root_type, NULL);
 
     /* release resources */
     flb_http_client_destroy(c);
@@ -1404,7 +1400,8 @@ static int flb_kube_network_init(struct flb_kube *ctx, struct flb_config *config
         if (!ctx->tls_ca_path && !ctx->tls_ca_file) {
             ctx->tls_ca_file  = flb_strdup(FLB_KUBE_CA);
         }
-        ctx->tls = flb_tls_create(ctx->tls_verify,
+        ctx->tls = flb_tls_create(FLB_TLS_CLIENT_MODE,
+                                  ctx->tls_verify,
                                   ctx->tls_debug,
                                   ctx->tls_vhost,
                                   ctx->tls_ca_path,
@@ -1430,7 +1427,7 @@ static int flb_kube_network_init(struct flb_kube *ctx, struct flb_config *config
     }
 
     /* Remove async flag from upstream */
-    ctx->upstream->flags &= ~(FLB_IO_ASYNC);
+    flb_stream_disable_async_mode(&ctx->upstream->base);
 
     return 0;
 }
@@ -1553,9 +1550,9 @@ int flb_kube_meta_get(struct flb_kube *ctx,
     }
 
     /* Check if we have some data associated to the cache key */
-    ret = flb_hash_get(ctx->hash_table,
-                       meta->cache_key, meta->cache_key_len,
-                       (void *) &hash_meta_buf, &hash_meta_size);
+    ret = flb_hash_table_get(ctx->hash_table,
+                             meta->cache_key, meta->cache_key_len,
+                             (void *) &hash_meta_buf, &hash_meta_size);
     if (ret == -1) {
         /* Retrieve API server meta and merge with local meta */
         ret = get_and_merge_meta(ctx, meta,
@@ -1566,9 +1563,9 @@ int flb_kube_meta_get(struct flb_kube *ctx,
             return 0;
         }
 
-        id = flb_hash_add(ctx->hash_table,
-                          meta->cache_key, meta->cache_key_len,
-                          tmp_hash_meta_buf, hash_meta_size);
+        id = flb_hash_table_add(ctx->hash_table,
+                                meta->cache_key, meta->cache_key_len,
+                                tmp_hash_meta_buf, hash_meta_size);
         if (id >= 0) {
             /*
              * Release the original buffer created on extract_meta() as a new
@@ -1576,8 +1573,8 @@ int flb_kube_meta_get(struct flb_kube *ctx,
              * the outgoing buffer and size.
              */
             flb_free(tmp_hash_meta_buf);
-            flb_hash_get_by_id(ctx->hash_table, id, meta->cache_key,
-                               &hash_meta_buf, &hash_meta_size);
+            flb_hash_table_get_by_id(ctx->hash_table, id, meta->cache_key,
+                                     &hash_meta_buf, &hash_meta_size);
         }
     }
 
