@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2022 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -26,6 +25,8 @@
 #include "systemd_config.h"
 #include "systemd_db.h"
 
+#include <ctype.h>
+
 /* msgpack helpers to pack unsigned ints (it takes care of endianness */
 #define pack_uint16(buf, d) _msgpack_store16(buf, (uint16_t) d)
 #define pack_uint32(buf, d) _msgpack_store32(buf, (uint32_t) d)
@@ -74,41 +75,37 @@ static int in_systemd_collect(struct flb_input_instance *ins,
 {
     int ret;
     int ret_j;
+    int i;
     int len;
     int entries = 0;
+    int skip_entries = 0;
     int rows = 0;
     time_t sec;
     long nsec;
-    uint8_t h;
     uint64_t usec;
     size_t length;
+    size_t threshold;
     const char *sep;
     const char *key;
     const char *val;
-    char *tmp;
+    char *buf = NULL;
 #ifdef FLB_HAVE_SQLDB
     char *cursor = NULL;
 #endif
     char *tag = NULL;
     char new_tag[PATH_MAX];
-    char last_tag[PATH_MAX];
+    char last_tag[PATH_MAX] = {0};
     size_t tag_len;
     size_t last_tag_len = 0;
-    off_t off;
     const void *data;
     struct flb_systemd_config *ctx = in_context;
     struct flb_time tm;
-    msgpack_sbuffer mp_sbuf;
-    msgpack_packer mp_pck;
 
     /* Restricted by mem_buf_limit */
     if (flb_input_buf_paused(ins) == FLB_TRUE) {
         return FLB_SYSTEMD_BUSY;
     }
 
-    msgpack_sbuffer_init(&mp_sbuf);
-    msgpack_packer_init(&mp_pck, &mp_sbuf, msgpack_sbuffer_write);
-
     /*
      * if there are not pending records from a previous round, likely we got
      * some changes in the journal, otherwise go ahead and continue reading
@@ -125,6 +122,17 @@ static int in_systemd_collect(struct flb_input_instance *ins,
         }
     }
 
+    if (ctx->lowercase == FLB_TRUE) {
+        ret = sd_journal_get_data_threshold(ctx->j, &threshold);
+        if (ret != 0) {
+            flb_plg_error(ctx->ins,
+                          "error setting up systemd data. "
+                          "sd_journal_get_data_threshold() return value '%i'",
+                          ret);
+            return FLB_SYSTEMD_ERROR;
+        }
+    }
+
     while ((ret_j = sd_journal_next(ctx->j)) > 0) {
         /* If the tag is composed dynamically, gather the Systemd Unit name */
         if (ctx->dynamic_tag) {
@@ -171,38 +179,30 @@ static int in_systemd_collect(struct flb_input_instance *ins,
          * so a new msgpack buffer is required. We ingest the data and prepare
          * a new buffer.
          */
-        if (mp_sbuf.size > 0 &&
-            ((last_tag_len != tag_len) || (strncmp(last_tag, tag, tag_len) != 0))) {
-            flb_input_chunk_append_raw(ctx->ins,
-                                       last_tag, last_tag_len,
-                                       mp_sbuf.data,
-                                       mp_sbuf.size);
-            msgpack_sbuffer_destroy(&mp_sbuf);
-            msgpack_sbuffer_init(&mp_sbuf);
+        if (ctx->log_encoder->output_length > 0 &&
+            ((last_tag_len != tag_len) ||
+             (strncmp(last_tag, tag, tag_len) != 0))) {
+            flb_input_log_append(ctx->ins,
+                                 last_tag, last_tag_len,
+                                 ctx->log_encoder->output_buffer,
+                                 ctx->log_encoder->output_length);
+
+            flb_log_event_encoder_reset(ctx->log_encoder);
 
             strncpy(last_tag, tag, tag_len);
             last_tag_len = tag_len;
         }
 
-        /* Prepare buffer and write map content */
-        msgpack_pack_array(&mp_pck, 2);
-        flb_time_append_to_msgpack(&tm, &mp_pck, 0);
 
-        /*
-         * Save the current size/position of the buffer since this is
-         * where the Map header will be stored.
-         */
-        off = mp_sbuf.size;
+        ret = flb_log_event_encoder_begin_record(ctx->log_encoder);
 
-        /*
-         * Register the maximum fields allowed per entry in the map. With
-         * this approach we can ingest all the fields and then just adjust
-         * the map size if required.
-         */
-        msgpack_pack_map(&mp_pck, ctx->max_fields);
+        if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+            ret = flb_log_event_encoder_set_timestamp(ctx->log_encoder, &tm);
+        }
 
         /* Pack every field in the entry */
         entries = 0;
+        skip_entries = 0;
         while (sd_journal_enumerate_data(ctx->j, &data, &length) > 0 &&
                entries < ctx->max_fields) {
             key = (const char *) data;
@@ -210,56 +210,82 @@ static int in_systemd_collect(struct flb_input_instance *ins,
                 key++;
                 length--;
             }
+
             sep = strchr(key, '=');
+            if (sep == NULL) {
+                skip_entries++;
+                continue;
+            }
+
             len = (sep - key);
-            msgpack_pack_str(&mp_pck, len);
-            msgpack_pack_str_body(&mp_pck, key, len);
+
+            if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+                ret = flb_log_event_encoder_append_body_string_length(
+                        ctx->log_encoder, len);
+            }
+
+            if (ctx->lowercase == FLB_TRUE) {
+                /*
+                 * Ensure buf to have enough space for the key because the libsystemd
+                 * might return larger data than the threshold.
+                 */
+                if (buf == NULL) {
+                    buf = flb_sds_create_len(NULL, threshold);
+                }
+                if (flb_sds_alloc(buf) < len) {
+                    buf = flb_sds_increase(buf, len - flb_sds_alloc(buf));
+                }
+                for (i = 0; i < len; i++) {
+                    buf[i] = tolower(key[i]);
+                }
+
+                if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+                    ret = flb_log_event_encoder_append_body_string_body(
+                            ctx->log_encoder, buf, len);
+                }
+            }
+            else {
+                if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+                    ret = flb_log_event_encoder_append_body_string_body(
+                            ctx->log_encoder, (char *) key, len);
+                }
+            }
 
             val = sep + 1;
             len = length - (sep - key) - 1;
-            msgpack_pack_str(&mp_pck, len);
-            msgpack_pack_str_body(&mp_pck, val, len);
+
+            if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+                ret = flb_log_event_encoder_append_body_string(
+                        ctx->log_encoder, (char *) val, len);
+            }
 
             entries++;
         }
         rows++;
-        if (entries == ctx->max_fields) {
-            flb_plg_debug(ctx->ins,
-                          "max number of fields is reached: %i; all other "
-                          "fields are discarded", ctx->max_fields);
+
+        if (skip_entries > 0) {
+            flb_plg_error(ctx->ins, "Skip %d broken entries", skip_entries);
         }
 
-        /*
-         * The fields were packed, now we need to adjust the msgpack map size
-         * to set the proper number of fields appended to the record.
-         */
-        tmp = mp_sbuf.data + off;
-        h = tmp[0];
-        if (h >> 4 == 0x8) {
-            *tmp = (uint8_t) 0x8 << 4 | ((uint8_t) entries);
-        }
-        else if (h == 0xde) {
-            tmp++;
-            pack_uint16(tmp, entries);
-        }
-        else if (h == 0xdf) {
-            tmp++;
-            pack_uint32(tmp, entries);
+        if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+            ret = flb_log_event_encoder_commit_record(ctx->log_encoder);
         }
 
         /*
          * Some journals can have too much data, pause if we have processed
          * more than 1MB. Journal will resume later.
          */
-        if (mp_sbuf.size > 1024000) {
-            flb_input_chunk_append_raw(ctx->ins,
-                                       tag, tag_len,
-                                       mp_sbuf.data,
-                                       mp_sbuf.size);
-            msgpack_sbuffer_destroy(&mp_sbuf);
-            msgpack_sbuffer_init(&mp_sbuf);
+        if (ctx->log_encoder->output_length > 1024000) {
+            flb_input_log_append(ctx->ins,
+                                 tag, tag_len,
+                                 ctx->log_encoder->output_buffer,
+                                 ctx->log_encoder->output_length);
+
+            flb_log_event_encoder_reset(ctx->log_encoder);
+
             strncpy(last_tag, tag, tag_len);
             last_tag_len = tag_len;
+
             break;
         }
 
@@ -268,6 +294,8 @@ static int in_systemd_collect(struct flb_input_instance *ins,
         }
     }
 
+    flb_sds_destroy(buf);
+
 #ifdef FLB_HAVE_SQLDB
     /* Save cursor */
     if (ctx->db) {
@@ -280,13 +308,14 @@ static int in_systemd_collect(struct flb_input_instance *ins,
 #endif
 
     /* Write any pending data into the buffer */
-    if (mp_sbuf.size > 0) {
-        flb_input_chunk_append_raw(ctx->ins,
-                                   tag, tag_len,
-                                   mp_sbuf.data,
-                                   mp_sbuf.size);
+    if (ctx->log_encoder->output_length > 0) {
+        flb_input_log_append(ctx->ins,
+                             tag, tag_len,
+                             ctx->log_encoder->output_buffer,
+                             ctx->log_encoder->output_length);
+
+        flb_log_event_encoder_reset(ctx->log_encoder);
     }
-    msgpack_sbuffer_destroy(&mp_sbuf);
 
     /* the journal is empty, no more records */
     if (ret_j == 0) {
@@ -454,6 +483,63 @@ static int in_systemd_exit(void *data, struct flb_config *config)
     return 0;
 }
 
+static struct flb_config_map config_map[] = {
+    {
+      FLB_CONFIG_MAP_STR, "path", (char *)NULL,
+      0, FLB_TRUE, offsetof(struct flb_systemd_config, path),
+      "Set the systemd journal path"
+    },
+    {
+      FLB_CONFIG_MAP_INT, "max_fields", FLB_SYSTEMD_MAX_FIELDS,
+      0, FLB_TRUE, offsetof(struct flb_systemd_config, max_fields),
+      "Set the maximum fields per notification"
+    },
+    {
+      FLB_CONFIG_MAP_INT, "max_entries", FLB_SYSTEMD_MAX_ENTRIES,
+      0, FLB_TRUE, offsetof(struct flb_systemd_config, max_entries),
+      "Set the maximum entries per notification"
+    },
+    {
+      FLB_CONFIG_MAP_STR, "systemd_filter_type", (char *)NULL,
+      0, FLB_TRUE, offsetof(struct flb_systemd_config, filter_type),
+      "Set the systemd filter type to either 'and' or 'or'"
+    },
+    {
+      FLB_CONFIG_MAP_STR, "systemd_filter", (char *)NULL,
+      FLB_CONFIG_MAP_MULT, FLB_TRUE, offsetof(struct flb_systemd_config, systemd_filters),
+      "Add a systemd filter, can be set multiple times"
+    },
+    {
+      FLB_CONFIG_MAP_BOOL, "read_from_tail", "false",
+      0, FLB_TRUE, offsetof(struct flb_systemd_config, read_from_tail),
+      "Read the journal from the end (tail)"
+    },
+    {
+      FLB_CONFIG_MAP_BOOL, "lowercase", "false",
+      0, FLB_TRUE, offsetof(struct flb_systemd_config, lowercase),
+      "Lowercase the fields"
+    },
+    {
+      FLB_CONFIG_MAP_BOOL, "strip_underscores", "false",
+      0, FLB_TRUE, offsetof(struct flb_systemd_config, strip_underscores),
+      "Strip undersecores from fields"
+    },
+#ifdef FLB_HAVE_SQLDB
+    {
+      FLB_CONFIG_MAP_STR, "db.sync", (char *)NULL,
+      0, FLB_TRUE, offsetof(struct flb_systemd_config, db_sync_mode),
+      "Set the database sync mode: extra, full, normal or off"
+    },
+    {
+      FLB_CONFIG_MAP_STR, "db", (char *)NULL,
+      0, FLB_TRUE, offsetof(struct flb_systemd_config, db_path),
+      "Set the database path"
+    },
+#endif /* FLB_HAVE_SQLDB */
+    /* EOF */
+    {0}
+};
+
 /* Plugin reference */
 struct flb_input_plugin in_systemd_plugin = {
     .name         = "systemd",
@@ -464,5 +550,6 @@ struct flb_input_plugin in_systemd_plugin = {
     .cb_pause     = in_systemd_pause,
     .cb_resume    = in_systemd_resume,
     .cb_exit      = in_systemd_exit,
+    .config_map   = config_map,
     .flags        = 0
 };
