@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2022 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -35,25 +34,25 @@
 #define FLB_STD_SCOPE     "https://www.googleapis.com/auth/logging.write"
 
 /* Stackdriver authorization URL */
-#define FLB_STD_AUTH_URL  "https://www.googleapis.com/oauth2/v4/token"
+#define FLB_STD_AUTH_URL  "https://oauth2.googleapis.com/token"
 
 /* Stackdriver Logging 'write' end-point */
 #define FLB_STD_WRITE_URI "/v2/entries:write"
-#define FLB_STD_WRITE_URL \
-    "https://logging.googleapis.com" FLB_STD_WRITE_URI
+#define FLB_STD_WRITE_URL "https://logging.googleapis.com" FLB_STD_WRITE_URI
 
 /* Timestamp format */
 #define FLB_STD_TIME_FMT  "%Y-%m-%dT%H:%M:%S"
 
 /* Default Resource type */
 #define FLB_SDS_RESOURCE_TYPE "global"
-
 #define OPERATION_FIELD_IN_JSON "logging.googleapis.com/operation"
 #define MONITORED_RESOURCE_KEY "logging.googleapis.com/monitored_resource"
 #define LOCAL_RESOURCE_ID_KEY "logging.googleapis.com/local_resource_id"
 #define DEFAULT_LABELS_KEY "logging.googleapis.com/labels"
 #define DEFAULT_SEVERITY_KEY "logging.googleapis.com/severity"
 #define DEFAULT_TRACE_KEY "logging.googleapis.com/trace"
+#define DEFAULT_SPAN_ID_KEY "logging.googleapis.com/spanId"
+#define DEFAULT_TRACE_SAMPLED_KEY "logging.googleapis.com/traceSampled"
 #define DEFAULT_LOG_NAME_KEY "logging.googleapis.com/logName"
 #define DEFAULT_INSERT_ID_KEY "logging.googleapis.com/insertId"
 #define SOURCELOCATION_FIELD_IN_JSON "logging.googleapis.com/sourceLocation"
@@ -64,6 +63,14 @@
 #define SOURCE_LOCATION_SIZE 37
 #define HTTP_REQUEST_KEY_SIZE 35
 
+/*
+ * Stackdriver implements a specific HTTP status code that is used internally in clients to keep
+ * track of networking errors that could happen before a successful HTTP request/response. For
+ * metrics counting purposes, every failed networking connection will use a 502 HTTP status code
+ * that can be used later to query the metrics by using labels with that value.
+ */
+#define STACKDRIVER_NET_ERROR  502
+
 #define K8S_CONTAINER "k8s_container"
 #define K8S_NODE      "k8s_node"
 #define K8S_POD       "k8s_pod"
@@ -79,6 +86,22 @@
 #define FLB_STACKDRIVER_FAILED_REQUESTS      1001   /* failed requests */
 #endif
 
+struct flb_stackdriver_oauth_credentials {
+    /* parsed credentials file */
+    flb_sds_t type;
+    flb_sds_t private_key_id;
+    flb_sds_t private_key;
+    flb_sds_t client_email;
+    flb_sds_t client_id;
+    flb_sds_t auth_uri;
+    flb_sds_t token_uri;
+};
+
+struct flb_stackdriver_env {
+    flb_sds_t creds_file;
+    flb_sds_t metadata_server;
+};
+
 struct flb_stackdriver {
     /* credentials */
     flb_sds_t credentials_file;
@@ -107,16 +130,28 @@ struct flb_stackdriver {
     flb_sds_t pod_name;
     flb_sds_t container_name;
     flb_sds_t node_name;
-    bool is_k8s_resource_type;
 
-    flb_sds_t labels_key;
     flb_sds_t local_resource_id;
     flb_sds_t tag_prefix;
+    /* shadow tag_prefix for safe deallocation */
+    flb_sds_t tag_prefix_k8s;
+
+    /* labels */
+    flb_sds_t labels_key;
+    struct mk_list *labels;
+    struct mk_list config_labels;
+
+    /* resource type flag */
+    int resource_type;
+
+    /* resource labels api */
+    struct mk_list *resource_labels;
+    struct mk_list resource_labels_kvs;
+    int should_skip_resource_labels_api;
 
     /* generic resources */
     flb_sds_t location;
     flb_sds_t namespace_id;
-    bool is_generic_resource_type;
 
     /* generic_node specific */
     flb_sds_t node_id;
@@ -125,11 +160,16 @@ struct flb_stackdriver {
     flb_sds_t job;
     flb_sds_t task_id;
 
+    /* Internal variable to reduce string comparisons */
+    int compress_gzip;
+
     /* other */
     flb_sds_t export_to_project_id;
     flb_sds_t resource;
     flb_sds_t severity_key;
     flb_sds_t trace_key;
+    flb_sds_t span_id_key;
+    flb_sds_t trace_sampled_key;
     flb_sds_t log_name_key;
     flb_sds_t http_request_key;
     int http_request_key_size;
@@ -144,6 +184,12 @@ struct flb_stackdriver {
     /* oauth2 context */
     struct flb_oauth2 *o;
 
+    /* parsed oauth2 credentials */
+    struct flb_stackdriver_oauth_credentials *creds;
+
+    /* environment variable settings */
+    struct flb_stackdriver_env *env;
+
     /* mutex for acquiring oauth tokens */
     pthread_mutex_t token_mutex;
 
@@ -157,6 +203,9 @@ struct flb_stackdriver {
     /* metrics */
     struct cmt_counter *cmt_successful_requests;
     struct cmt_counter *cmt_failed_requests;
+    struct cmt_counter *cmt_requests_total;
+    struct cmt_counter *cmt_proc_records_total;
+    struct cmt_counter *cmt_retried_records_total;
 #endif
 
     /* plugin instance */
