@@ -2,8 +2,7 @@
 
 /*  Fluent Bit Demo
  *  ===============
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2022 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -32,6 +31,8 @@
 #include <fluent-bit/flb_callback.h>
 #include <fluent-bit/flb_kv.h>
 #include <fluent-bit/flb_metrics.h>
+#include <fluent-bit/flb_upstream.h>
+#include <fluent-bit/flb_downstream.h>
 #include <fluent-bit/tls/flb_tls.h>
 
 #include <signal.h>
@@ -50,6 +51,12 @@ struct flb_aws_error_reporter *error_reporter;
 /* thread initializator */
 static pthread_once_t flb_lib_once = PTHREAD_ONCE_INIT;
 
+/* reference to the last 'flb_lib_ctx' context started through flb_start() */
+FLB_TLS_DEFINE(flb_ctx_t, flb_lib_active_context);
+
+/* reference to the last 'flb_cf' context started through flb_start() */
+FLB_TLS_DEFINE(struct flb_cf, flb_lib_active_cf_context);
+
 #ifdef FLB_SYSTEM_WINDOWS
 static inline int flb_socket_init_win32(void)
 {
@@ -118,8 +125,12 @@ void flb_init_env()
     flb_tls_init();
     flb_coro_init();
     flb_upstream_init();
+    flb_downstream_init();
     flb_output_prepare();
 
+    FLB_TLS_INIT(flb_lib_active_context);
+    FLB_TLS_INIT(flb_lib_active_cf_context);
+
     /* libraries */
     cmt_initialize();
 }
@@ -194,7 +205,7 @@ flb_ctx_t *flb_create()
                                   ctx->event_channel);
     if (ret != 0) {
         flb_error("[lib] could not create notification channels");
-        flb_config_exit(ctx->config);
+        flb_stop(ctx);
         flb_destroy(ctx);
         return NULL;
     }
@@ -224,7 +235,7 @@ void flb_destroy(flb_ctx_t *ctx)
     mk_event_loop_destroy(ctx->event_loop);
 
     /* cfg->is_running is set to false when flb_engine_shutdown has been invoked (event loop) */
-    if(ctx->config) {
+    if (ctx->config) {
         if (ctx->config->is_running == FLB_TRUE) {
             flb_engine_shutdown(ctx->config);
         }
@@ -597,7 +608,6 @@ int flb_lib_push(flb_ctx_t *ctx, int ffd, const void *data, size_t len)
     int ret;
     struct flb_input_instance *i_ins;
 
-
     if (ctx->status == FLB_LIB_NONE || ctx->status == FLB_LIB_ERROR) {
         flb_error("[lib] cannot push data, engine is not running");
         return -1;
@@ -623,12 +633,14 @@ static void flb_lib_worker(void *data)
     struct flb_config *config;
 
     config = ctx->config;
+    flb_context_set(ctx);
     mk_utils_worker_rename("flb-pipeline");
     ret = flb_engine_start(config);
     if (ret == -1) {
         flb_engine_failed(config);
         flb_engine_shutdown(config);
     }
+    config->exit_status_code = ret;
     ctx->status = FLB_LIB_NONE;
 }
 
@@ -641,8 +653,7 @@ double flb_time_now()
     return flb_time_to_double(&t);
 }
 
-/* Start the engine */
-int flb_start(flb_ctx_t *ctx)
+int static do_start(flb_ctx_t *ctx)
 {
     int fd;
     int bytes;
@@ -654,6 +665,11 @@ int flb_start(flb_ctx_t *ctx)
 
     pthread_once(&flb_lib_once, flb_init_env);
 
+    flb_debug("[lib] context set: %p", ctx);
+
+    /* set context as the last active one */
+
+    /* spawn worker thread */
     config = ctx->config;
     ret = mk_utils_worker_spawn(flb_lib_worker, ctx, &tid);
     if (ret == -1) {
@@ -667,7 +683,9 @@ int flb_start(flb_ctx_t *ctx)
         fd = event->fd;
         bytes = flb_pipe_r(fd, &val, sizeof(uint64_t));
         if (bytes <= 0) {
+#if defined(FLB_SYSTEM_MACOS)
             pthread_cancel(tid);
+#endif
             pthread_join(tid, NULL);
             ctx->status = FLB_LIB_ERROR;
             return -1;
@@ -680,15 +698,41 @@ int flb_start(flb_ctx_t *ctx)
         }
         else if (val == FLB_ENGINE_FAILED) {
             flb_error("[lib] backend failed");
+#if defined(FLB_SYSTEM_MACOS)
+            pthread_cancel(tid);
+#endif
             pthread_join(tid, NULL);
             ctx->status = FLB_LIB_ERROR;
             return -1;
         }
+        else {
+            flb_error("[lib] other error");
+        }
     }
 
     return 0;
 }
 
+/* Start the engine */
+int flb_start(flb_ctx_t *ctx)
+{
+    int ret;
+
+    ret = do_start(ctx);
+    if (ret == 0) {
+        /* set context as the last active one */
+        flb_context_set(ctx);
+    }
+
+    return ret;
+}
+
+/* Start the engine without setting the global context */
+int flb_start_trace(flb_ctx_t *ctx)
+{
+    return do_start(ctx);
+}
+
 int flb_loop(flb_ctx_t *ctx)
 {
     while (ctx->status == FLB_LIB_OK) {
@@ -703,7 +747,20 @@ int flb_stop(flb_ctx_t *ctx)
     int ret;
     pthread_t tid;
 
+    flb_debug("[lib] ctx stop address: %p, config context=%p\n", ctx, ctx->config);
+
+    tid = ctx->config->worker;
+
     if (ctx->status == FLB_LIB_NONE || ctx->status == FLB_LIB_ERROR) {
+        /*
+         * There is a chance the worker thread is still active while
+         * the service exited for some reason (plugin action). Always
+         * wait and double check that the child thread is not running.
+         */
+#if defined(FLB_SYSTEM_MACOS)
+        pthread_cancel(tid);
+#endif
+        pthread_join(tid, NULL);
         return 0;
     }
 
@@ -717,10 +774,42 @@ int flb_stop(flb_ctx_t *ctx)
 
     flb_debug("[lib] sending STOP signal to the engine");
 
-    tid = ctx->config->worker;
     flb_engine_exit(ctx->config);
+#if defined(FLB_SYSTEM_MACOS)
+    pthread_cancel(tid);
+#endif
     ret = pthread_join(tid, NULL);
+    if (ret != 0) {
+        flb_errno();
+    }
     flb_debug("[lib] Fluent Bit engine stopped");
 
     return ret;
 }
+
+
+void flb_context_set(flb_ctx_t *ctx)
+{
+    FLB_TLS_SET(flb_lib_active_context, ctx);
+}
+
+flb_ctx_t *flb_context_get()
+{
+    flb_ctx_t *ctx;
+
+    ctx = FLB_TLS_GET(flb_lib_active_context);
+    return ctx;
+}
+
+void flb_cf_context_set(struct flb_cf *cf)
+{
+    FLB_TLS_SET(flb_lib_active_cf_context, cf);
+}
+
+struct flb_cf *flb_cf_context_get()
+{
+    struct flb_cf *cf;
+
+    cf = FLB_TLS_GET(flb_lib_active_cf_context);
+    return cf;
+}
