@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2022 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -31,10 +30,15 @@
 #include <fluent-bit/flb_strptime.h>
 #include <fluent-bit/flb_env.h>
 #include <fluent-bit/flb_str.h>
+#include <fluent-bit/flb_kv.h>
+#include <fluent-bit/flb_config_format.h>
 #include <fluent-bit/multiline/flb_ml.h>
 #include <fluent-bit/multiline/flb_ml_parser.h>
 #include <fluent-bit/multiline/flb_ml_rule.h>
 
+#include <cfl/cfl.h>
+#include <cfl/cfl_kvlist.h>
+
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <limits.h>
@@ -126,11 +130,13 @@ static void flb_interim_parser_destroy(struct flb_parser *parser)
     flb_free(parser->name);
     if (parser->time_fmt) {
         flb_free(parser->time_fmt);
-        flb_free(parser->time_fmt_full);
     }
     if (parser->time_fmt_year) {
         flb_free(parser->time_fmt_year);
     }
+    if (parser->time_fmt_full) {
+        flb_free(parser->time_fmt_full);
+    }
     if (parser->time_key) {
         flb_free(parser->time_key);
     }
@@ -141,10 +147,12 @@ static void flb_interim_parser_destroy(struct flb_parser *parser)
 
 struct flb_parser *flb_parser_create(const char *name, const char *format,
                                      const char *p_regex,
+                                     int skip_empty,
                                      const char *time_fmt, const char *time_key,
                                      const char *time_offset,
                                      int time_keep,
                                      int time_strict,
+                                     int logfmt_no_bare_keys,
                                      struct flb_parser_types *types,
                                      int types_len,
                                      struct mk_list *decoders,
@@ -164,7 +172,7 @@ struct flb_parser *flb_parser_create(const char *name, const char *format,
     /* Iterate current parsers and make sure the new one don't exists */
     mk_list_foreach(head, &config->parsers) {
         p = mk_list_entry(head, struct flb_parser, _head);
-        if (strcmp(p->name, name) == 0) {
+        if (p->name && strcmp(p->name, name) == 0) {
             flb_error("[parser] parser named '%s' already exists, skip.",
                       name);
             return NULL;
@@ -187,14 +195,15 @@ struct flb_parser *flb_parser_create(const char *name, const char *format,
     else if (strcasecmp(format, "json") == 0) {
         p->type = FLB_PARSER_JSON;
     }
-    else if (strcmp(format, "ltsv") == 0) {
+    else if (strcasecmp(format, "ltsv") == 0) {
         p->type = FLB_PARSER_LTSV;
     }
-    else if (strcmp(format, "logfmt") == 0) {
+    else if (strcasecmp(format, "logfmt") == 0) {
         p->type = FLB_PARSER_LOGFMT;
     }
     else {
         flb_error("[parser:%s] Invalid format %s", name, format);
+        mk_list_del(&p->_head);
         flb_free(p);
         return NULL;
     }
@@ -202,6 +211,7 @@ struct flb_parser *flb_parser_create(const char *name, const char *format,
     if (p->type == FLB_PARSER_REGEX) {
         if (!p_regex) {
             flb_error("[parser:%s] Invalid regex pattern", name);
+            mk_list_del(&p->_head);
             flb_free(p);
             return NULL;
         }
@@ -209,10 +219,12 @@ struct flb_parser *flb_parser_create(const char *name, const char *format,
         regex = flb_regex_create(p_regex);
         if (!regex) {
             flb_error("[parser:%s] Invalid regex pattern %s", name, p_regex);
+            mk_list_del(&p->_head);
             flb_free(p);
             return NULL;
         }
         p->regex = regex;
+        p->skip_empty = skip_empty;
         p->p_regex = flb_strdup(p_regex);
     }
 
@@ -220,7 +232,17 @@ struct flb_parser *flb_parser_create(const char *name, const char *format,
 
     if (time_fmt) {
         p->time_fmt_full = flb_strdup(time_fmt);
+        if (!p->time_fmt_full) {
+            flb_error("[parser:%s] could not duplicate time fmt full", name);
+            flb_interim_parser_destroy(p);
+            return NULL;
+        }
         p->time_fmt = flb_strdup(time_fmt);
+        if (!p->time_fmt) {
+            flb_error("[parser:%s] could not duplicate time fmt", name);
+            flb_interim_parser_destroy(p);
+            return NULL;
+        }
 
         /* Check if the format is considering the year */
         if (strstr(p->time_fmt, "%Y") || strstr(p->time_fmt, "%y")) {
@@ -309,6 +331,7 @@ struct flb_parser *flb_parser_create(const char *name, const char *format,
 
     p->time_keep = time_keep;
     p->time_strict = time_strict;
+    p->logfmt_no_bare_keys = logfmt_no_bare_keys;
     p->types = types;
     p->types_len = types_len;
     return p;
@@ -317,6 +340,7 @@ struct flb_parser *flb_parser_create(const char *name, const char *format,
 void flb_parser_destroy(struct flb_parser *parser)
 {
     int i = 0;
+
     if (parser->type == FLB_PARSER_REGEX) {
         flb_regex_destroy(parser->regex);
         flb_free(parser->p_regex);
@@ -418,37 +442,40 @@ static int proc_types_str(const char *types_str, struct flb_parser_types **types
     return i;
 }
 
-static flb_sds_t get_parser_key(char *key,
-                                struct flb_config *config,
-                                struct mk_rconf_section *section)
+static flb_sds_t get_parser_key(struct flb_config *config,
+                                struct flb_cf *cf, struct flb_cf_section *s,
+                                char *key)
+
 {
-    char *tmp;
+    flb_sds_t tmp;
     flb_sds_t val;
 
-    tmp = mk_rconf_section_get_key(section, key, MK_RCONF_STR);
+    tmp = flb_cf_section_property_get_string(cf, s, key);
     if (!tmp) {
         return NULL;
     }
 
     val = flb_env_var_translate(config->env, tmp);
-    flb_free(tmp);
-
     if (!val) {
+        flb_sds_destroy(tmp);
         return NULL;
     }
 
     if (flb_sds_len(val) == 0) {
         flb_sds_destroy(val);
+        flb_sds_destroy(tmp);
         return NULL;
     }
 
+    flb_sds_destroy(tmp);
     return val;
 }
 
-/* Config file: read [PARSER] definitions */
-static int parser_conf_file(const char *cfg, struct mk_rconf *fconf,
+/* Config file: read 'parser' definitions */
+static int parser_conf_file(const char *cfg, struct flb_cf *cf,
                             struct flb_config *config)
 {
+    int i = 0;
     flb_sds_t name;
     flb_sds_t format;
     flb_sds_t regex;
@@ -457,16 +484,18 @@ static int parser_conf_file(const char *cfg, struct mk_rconf *fconf,
     flb_sds_t time_offset;
     flb_sds_t types_str;
     flb_sds_t tmp_str;
+    int skip_empty;
     int time_keep;
     int time_strict;
+    int logfmt_no_bare_keys;
     int types_len;
     struct mk_list *head;
     struct mk_list *decoders = NULL;
-    struct mk_rconf_section *section;
+    struct flb_cf_section *s;
     struct flb_parser_types *types = NULL;
 
-    /* Read all [PARSER] sections */
-    mk_list_foreach(head, &fconf->sections) {
+    /* Read all 'parser' sections */
+    mk_list_foreach(head, &cf->parsers) {
         name = NULL;
         format = NULL;
         regex = NULL;
@@ -476,60 +505,75 @@ static int parser_conf_file(const char *cfg, struct mk_rconf *fconf,
         types_str = NULL;
         tmp_str = NULL;
 
-        section = mk_list_entry(head, struct mk_rconf_section, _head);
-        if (strcasecmp(section->name, "PARSER") != 0) {
-            continue;
-        }
+        /* retrieve the section context */
+        s = mk_list_entry(head, struct flb_cf_section, _head_section);
 
-        /* Name */
-        name = get_parser_key("Name", config, section);
+        /* name */
+        name = get_parser_key(config, cf, s, "name");
         if (!name) {
             flb_error("[parser] no parser 'name' found in file '%s'", cfg);
-            goto fconf_error;
+            goto fconf_early_error;
         }
 
-        /* Format */
-        format = get_parser_key("Format", config, section);
+        /* format */
+        format = get_parser_key(config, cf, s, "format");
         if (!format) {
             flb_error("[parser] no parser 'format' found for '%s' in file '%s'",
                       name, cfg);
-            goto fconf_error;
+            goto fconf_early_error;
         }
 
-        /* Regex (if 'format' == 'regex') */
-        regex = get_parser_key("Regex", config, section);
+        /* regex (if 'format' == 'regex') */
+        regex = get_parser_key(config, cf, s, "regex");
         if (!regex && strcmp(format, "regex") == 0) {
-            flb_error("[parser] no parser 'regex' found for '%s' in file '%s", name, cfg);
-            goto fconf_error;
+            flb_error("[parser] no parser 'regex' found for '%s' in file '%s",
+                      name, cfg);
+            goto fconf_early_error;
+        }
+        
+        /* skip_empty_values */
+        skip_empty = FLB_TRUE;
+        tmp_str = get_parser_key(config, cf, s, "skip_empty_values");
+        if (tmp_str) {
+            skip_empty = flb_utils_bool(tmp_str);
+            flb_sds_destroy(tmp_str);
         }
 
-        /* Time_Format */
-        time_fmt = get_parser_key("Time_Format", config, section);
+        /* time_format */
+        time_fmt = get_parser_key(config, cf, s, "time_format");
 
-        /* Time_Key */
-        time_key = get_parser_key("Time_Key", config, section);
+        /* time_key */
+        time_key = get_parser_key(config, cf, s, "time_key");
 
-        /* Time_Keep */
+        /* time_keep */
         time_keep = FLB_FALSE;
-        tmp_str = get_parser_key("Time_Keep", config, section);
+        tmp_str = get_parser_key(config, cf, s, "time_keep");
         if (tmp_str) {
             time_keep = flb_utils_bool(tmp_str);
             flb_sds_destroy(tmp_str);
         }
 
-        /* Time_Strict */
+        /* time_strict */
         time_strict = FLB_TRUE;
-        tmp_str = get_parser_key("Time_Strict", config, section);
+        tmp_str = get_parser_key(config, cf, s, "time_strict");
         if (tmp_str) {
             time_strict = flb_utils_bool(tmp_str);
             flb_sds_destroy(tmp_str);
         }
 
-        /* Time_Offset (UTC offset) */
-        time_offset = get_parser_key("Time_Offset", config, section);
+        /* time_offset (UTC offset) */
+        time_offset = get_parser_key(config, cf, s, "time_offset");
 
-        /* Types */
-        types_str = get_parser_key("Types", config, section);
+        /* logfmt_no_bare_keys */
+        logfmt_no_bare_keys = FLB_FALSE;
+        tmp_str = get_parser_key(config, cf, s, "logfmt_no_bare_keys");
+        if (tmp_str) {
+            logfmt_no_bare_keys = flb_utils_bool(tmp_str);
+            flb_sds_destroy(tmp_str);
+        }
+
+        /* types */
+        types_str = get_parser_key(config, cf, s, "types");
         if (types_str) {
             types_len = proc_types_str(types_str, &types);
         }
@@ -538,12 +582,12 @@ static int parser_conf_file(const char *cfg, struct mk_rconf *fconf,
         }
 
         /* Decoders */
-        decoders = flb_parser_decoder_list_create(section);
+        decoders = flb_parser_decoder_list_create(s);
 
         /* Create the parser context */
-        if (!flb_parser_create(name, format, regex,
+        if (!flb_parser_create(name, format, regex, skip_empty,
                                time_fmt, time_key, time_offset, time_keep, time_strict,
-                               types, types_len, decoders, config)) {
+                               logfmt_no_bare_keys, types, types_len, decoders, config)) {
             goto fconf_error;
         }
 
@@ -572,6 +616,19 @@ static int parser_conf_file(const char *cfg, struct mk_rconf *fconf,
 
     return 0;
 
+ /* Use early exit before call to flb_parser_create */
+ fconf_early_error:
+    if (name) {
+        flb_sds_destroy(name);
+    }
+    if (format) {
+        flb_sds_destroy(format);
+    }
+    if (regex) {
+        flb_sds_destroy(regex);
+    }
+    return -1;
+
  fconf_error:
     flb_sds_destroy(name);
     flb_sds_destroy(format);
@@ -590,6 +647,14 @@ static int parser_conf_file(const char *cfg, struct mk_rconf *fconf,
     if (types_str) {
         flb_sds_destroy(types_str);
     }
+    if (types_len) {
+        for (i=0; i<types_len; i++){
+            if (types[i].key != NULL) {
+                flb_free(types[i].key);
+            }
+        }
+        flb_free(types);
+    }
     if (decoders) {
         flb_parser_decoder_list_destroy(decoders);
     }
@@ -597,20 +662,20 @@ static int parser_conf_file(const char *cfg, struct mk_rconf *fconf,
 }
 
 static int multiline_load_regex_rules(struct flb_ml_parser *ml_parser,
-                                      struct mk_rconf_section *section,
+                                      struct flb_cf_section *section,
                                       struct flb_config *config)
 {
     int ret;
     char *to_state = NULL;
     struct mk_list list;
-    struct mk_list *head;
-    struct mk_rconf_entry *entry;
+    struct cfl_list *head;
+    struct cfl_kvpair *entry;
     struct flb_slist_entry *from_state;
     struct flb_slist_entry *regex_pattern;
     struct flb_slist_entry *tmp;
 
-    mk_list_foreach(head, &section->entries) {
-        entry = mk_list_entry(head, struct mk_rconf_entry, _head);
+    cfl_list_foreach(head, &section->properties->list) {
+        entry = cfl_list_entry(head, struct cfl_kvpair, _head);
 
         /* only process 'rule' keys */
         if (strcasecmp(entry->key, "rule") != 0) {
@@ -618,7 +683,7 @@ static int multiline_load_regex_rules(struct flb_ml_parser *ml_parser,
         }
 
         mk_list_init(&list);
-        ret = flb_slist_split_tokens(&list, entry->val, 3);
+        ret = flb_slist_split_tokens(&list, entry->val->data.as_string, 3);
         if (ret == -1) {
             flb_error("[multiline parser: %s] invalid section on key '%s'",
                       ml_parser->name, entry->key);
@@ -677,8 +742,8 @@ static int multiline_load_regex_rules(struct flb_ml_parser *ml_parser,
 }
 
 
-/* Config file: read [MULTILINE_PARSER] definitions */
-static int multiline_parser_conf_file(const char *cfg, struct mk_rconf *fconf,
+/* config file: read 'multiline_parser' sections */
+static int multiline_parser_conf_file(const char *cfg, struct flb_cf *cf,
                                       struct flb_config *config)
 {
     int ret;
@@ -692,13 +757,13 @@ static int multiline_parser_conf_file(const char *cfg, struct mk_rconf *fconf,
     flb_sds_t parser;
     flb_sds_t tmp;
     int flush_timeout;
-    struct flb_parser *parser_ctx;
+    struct flb_parser *parser_ctx = NULL;
     struct mk_list *head;
-    struct mk_rconf_section *section;
+    struct flb_cf_section *s;
     struct flb_ml_parser *ml_parser;
 
-    /* Read all [PARSER] sections */
-    mk_list_foreach(head, &fconf->sections) {
+    /* read all 'multiline_parser' sections */
+    mk_list_foreach(head, &cf->multiline_parsers) {
         ml_parser = NULL;
         name = NULL;
         type = -1;
@@ -711,20 +776,17 @@ static int multiline_parser_conf_file(const char *cfg, struct mk_rconf *fconf,
         flush_timeout = -1;
         tmp = NULL;
 
-        section = mk_list_entry(head, struct mk_rconf_section, _head);
-        if (strcasecmp(section->name, "MULTILINE_PARSER") != 0) {
-            continue;
-        }
+        s = mk_list_entry(head, struct flb_cf_section, _head_section);
 
         /* name */
-        name = get_parser_key("name", config, section);
+        name = get_parser_key(config, cf, s, "name");
         if (!name) {
             flb_error("[multiline_parser] no 'name' defined in file '%s'", cfg);
             goto fconf_error;
         }
 
         /* type */
-        tmp = get_parser_key("type", config, section);
+        tmp = get_parser_key(config, cf, s, "type");
         if (!tmp) {
             flb_error("[multiline_parser] no 'type' defined in file '%s'", cfg);
             goto fconf_error;
@@ -733,36 +795,35 @@ static int multiline_parser_conf_file(const char *cfg, struct mk_rconf *fconf,
             type = flb_ml_type_lookup(tmp);
             if (type == -1) {
                 flb_error("[multiline_parser] invalid type '%s'", tmp);
-                flb_sds_destroy(tmp);
                 goto fconf_error;
             }
             flb_sds_destroy(tmp);
         }
 
         /* match_string */
-        match_string = get_parser_key("match_string", config, section);
+        match_string = get_parser_key(config, cf, s, "match_string");
 
         /* negate */
-        tmp = get_parser_key("negate", config, section);
+        tmp = get_parser_key(config, cf, s, "negate");
         if (tmp) {
             negate = flb_utils_bool(tmp);
             flb_sds_destroy(tmp);
         }
 
         /* key_content */
-        key_content = get_parser_key("key_content", config, section);
+        key_content = get_parser_key(config, cf, s, "key_content");
 
         /* key_pattern */
-        key_pattern = get_parser_key("key_pattern", config, section);
+        key_pattern = get_parser_key(config, cf, s, "key_pattern");
 
         /* key_group */
-        key_group = get_parser_key("key_group", config, section);
+        key_group = get_parser_key(config, cf, s, "key_group");
 
         /* parser */
-        parser = get_parser_key("parser", config, section);
+        parser = get_parser_key(config, cf, s, "parser");
 
         /* flush_timeout */
-        tmp = get_parser_key("flush_timeout", config, section);
+        tmp = get_parser_key(config, cf, s, "flush_timeout");
         if (tmp) {
             flush_timeout = atoi(tmp);
         }
@@ -780,7 +841,7 @@ static int multiline_parser_conf_file(const char *cfg, struct mk_rconf *fconf,
 
         /* if type is regex, process rules */
         if (type == FLB_ML_REGEX) {
-            ret = multiline_load_regex_rules(ml_parser, section, config);
+            ret = multiline_load_regex_rules(ml_parser, s, config);
             if (ret != 0) {
                 goto fconf_error;
             }
@@ -812,16 +873,11 @@ static int multiline_parser_conf_file(const char *cfg, struct mk_rconf *fconf,
     return -1;
 }
 
-/* Load parsers from a configuration file */
-int flb_parser_conf_file(const char *file, struct flb_config *config)
+int flb_parser_conf_file_stat(const char *file, struct flb_config *config)
 {
     int ret;
-    char tmp[PATH_MAX + 1];
-    const char *cfg = NULL;
-    struct mk_rconf *fconf;
     struct stat st;
 
-#ifndef FLB_HAVE_STATIC_CONF
     ret = stat(file, &st);
     if (ret == -1 && errno == ENOENT) {
         /* Try to resolve the real path (if exists) */
@@ -831,38 +887,62 @@ int flb_parser_conf_file(const char *file, struct flb_config *config)
         }
 
         if (config->conf_path) {
-            snprintf(tmp, PATH_MAX, "%s%s", config->conf_path, file);
-            cfg = tmp;
+            /* Handle as special case here. */
+            return -2;
         }
+
+        return -1;
+    }
+
+    return 0;
+}
+
+/* Load parsers from a configuration file */
+int flb_parser_conf_file(const char *file, struct flb_config *config)
+{
+    int ret;
+    char tmp[PATH_MAX + 1];
+    char *cfg = NULL;
+    struct flb_cf *cf = NULL;
+
+#ifndef FLB_HAVE_STATIC_CONF
+    ret = flb_parser_conf_file_stat(file, config);
+    if (ret == -1) {
+        return -1;
+    }
+    else if (ret == -2) {
+        snprintf(tmp, PATH_MAX, "%s%s", config->conf_path, file);
+        cfg = tmp;
     }
     else {
-        cfg = file;
+        cfg = (char *) file;
     }
 
-    fconf = mk_rconf_open(cfg);
+    cf = flb_cf_create_from_file(NULL, cfg);
 #else
-    fconf = flb_config_static_open(file);
+    cf = flb_config_static_open(file);
 #endif
 
-    if (!fconf) {
+    if (!cf) {
         return -1;
     }
 
-    /* process [PARSER]'s sections */
-    ret = parser_conf_file(cfg, fconf, config);
+    /* process 'parser' sections */
+    ret = parser_conf_file(cfg, cf, config);
     if (ret == -1) {
-        mk_rconf_free(fconf);
+        flb_cf_destroy(cf);
         return -1;
     }
 
-    /* processs [MULTILINE_PARSER]'s sections */
-    ret = multiline_parser_conf_file(cfg, fconf, config);
+    /* processs 'multiline_parser' sections */
+    ret = multiline_parser_conf_file(cfg, cf, config);
     if (ret == -1) {
-        mk_rconf_free(fconf);
+        flb_cf_destroy(cf);
         return -1;
     }
 
-    mk_rconf_free(fconf);
+    /* link the 'cf parser' context to the config list */
+    mk_list_add(&cf->_head, &config->cf_parsers_list);
     return 0;
 }
 
@@ -871,8 +951,15 @@ struct flb_parser *flb_parser_get(const char *name, struct flb_config *config)
     struct mk_list *head;
     struct flb_parser *parser;
 
+    if (config == NULL || mk_list_size(&config->parsers) <= 0) {
+        return NULL;
+    }
+
     mk_list_foreach(head, &config->parsers) {
         parser = mk_list_entry(head, struct flb_parser, _head);
+        if (parser == NULL || parser->name == NULL) {
+            continue;
+        }
         if (strcmp(parser->name, name) == 0) {
             return parser;
         }
@@ -927,6 +1014,12 @@ int flb_parser_tzone_offset(const char *str, int len, int *tmdiff)
         return -1;
     }
 
+    /* Ensure there is enough data */
+    if (len < 4) {
+        *tmdiff = 0;
+        return -1;
+    }
+
     /* Negative value ? */
     neg = (*p++ == '-');
 
@@ -936,6 +1029,11 @@ int flb_parser_tzone_offset(const char *str, int len, int *tmdiff)
     /* Gather hours and minutes */
     hour = ((p[0] - '0') * 10) + (p[1] - '0');
     if (end - p == 5 && p[2] == ':') {
+        /* Ensure there is enough data */
+        if (len < 5) {
+            *tmdiff = 0;
+            return -1;
+        }
         min = ((p[3] - '0') * 10) + (p[4] - '0');
     }
     else {
@@ -988,7 +1086,7 @@ static int parse_subseconds(char *str, int len, double *subsec)
 int flb_parser_time_lookup(const char *time_str, size_t tsize,
                            time_t now,
                            struct flb_parser *parser,
-                           struct tm *tm, double *ns)
+                           struct flb_tm *tm, double *ns)
 {
     int ret;
     time_t time_now;
@@ -1031,8 +1129,8 @@ int flb_parser_time_lookup(const char *time_str, size_t tsize,
         gmtime_r(&time_now, &tmy);
 
         /* Make the timestamp default to today */
-        tm->tm_mon = tmy.tm_mon;
-        tm->tm_mday = tmy.tm_mday;
+        tm->tm.tm_mon = tmy.tm_mon;
+        tm->tm.tm_mday = tmy.tm_mday;
 
         uint64_t t = tmy.tm_year + 1900;
 
@@ -1068,10 +1166,10 @@ int flb_parser_time_lookup(const char *time_str, size_t tsize,
 
     if (p == NULL) {
         if (parser->time_strict) {
-            flb_error("[parser] cannot parse '%.*s'", tsize, time_str);
+            flb_error("[parser] cannot parse '%.*s'", (int)tsize, time_str);
             return -1;
         }
-        flb_debug("[parser] non-exact match '%.*s'", tsize, time_str);
+        flb_debug("[parser] non-exact match '%.*s'", (int)tsize, time_str);
         return 0;
     }
 
@@ -1079,10 +1177,10 @@ int flb_parser_time_lookup(const char *time_str, size_t tsize,
         ret = parse_subseconds(p, time_len - (p - time_ptr), ns);
         if (ret < 0) {
             if (parser->time_strict) {
-                flb_error("[parser] cannot parse %%L for '%.*s'", tsize, time_str);
+                flb_error("[parser] cannot parse %%L for '%.*s'", (int)tsize, time_str);
                 return -1;
             }
-            flb_debug("[parser] non-exact match on %%L '%.*s'", tsize, time_str);
+            flb_debug("[parser] non-exact match on %%L '%.*s'", (int)tsize, time_str);
             return 0;
         }
         p += ret;
@@ -1091,19 +1189,17 @@ int flb_parser_time_lookup(const char *time_str, size_t tsize,
         p = flb_strptime(p, parser->time_frac_secs, tm);
         if (p == NULL) {
             if (parser->time_strict) {
-                flb_error("[parser] cannot parse '%.*s' after %%L", tsize, time_str);
+                flb_error("[parser] cannot parse '%.*s' after %%L", (int)tsize, time_str);
                 return -1;
             }
-            flb_debug("[parser] non-exact match after %%L '%.*s'", tsize, time_str);
+            flb_debug("[parser] non-exact match after %%L '%.*s'", (int)tsize, time_str);
             return 0;
         }
     }
 
-#ifdef FLB_HAVE_GMTOFF
     if (parser->time_with_tz == FLB_FALSE) {
-        tm->tm_gmtoff = parser->time_offset;
+        flb_tm_gmtoff(tm) = parser->time_offset;
     }
-#endif
 
     return 0;
 }
