@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2020 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2022 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -21,6 +20,7 @@
 #include <fluent-bit/flb_info.h>
 #include <fluent-bit/flb_compat.h>
 #include <fluent-bit/tls/flb_tls.h>
+#include <fluent-bit/tls/flb_tls_info.h>
 
 #include <openssl/ssl.h>
 #include <openssl/err.h>
@@ -34,25 +34,18 @@
  */
 #define OPENSSL_1_1_0 0x010100000L
 
-/*
- * RHEL-family distrbutions do not provide system certificates in
- * a format that OpenSSL's CAPath can read, but do provide a single
- * packed cert in /etc/certs.
- *
- * Use the bundled cert as the default trusted CA.
- */
-#define RHEL_DEFAULT_CA "/etc/ssl/certs/ca-bundle.crt"
-
 /* OpenSSL library context */
 struct tls_context {
     int debug_level;
     SSL_CTX *ctx;
+    int mode;
     pthread_mutex_t mutex;
 };
 
 struct tls_session {
     SSL *ssl;
     int fd;
+    int continuation_flag;
     struct tls_context *parent;    /* parent struct tls_context ref */
 };
 
@@ -134,22 +127,69 @@ static void tls_context_destroy(void *ctx_backend)
     flb_free(ctx);
 }
 
+#ifdef _MSC_VER
+static int windows_load_system_certificates(struct tls_context *ctx)
+{
+    int ret;
+    HANDLE win_store;
+    PCCERT_CONTEXT win_cert = NULL;
+    const unsigned char *win_cert_data;
+    X509_STORE *ossl_store = SSL_CTX_get_cert_store(ctx->ctx);
+    X509 *ossl_cert;
+
+    win_store = CertOpenSystemStoreA(0, "Root");
+    if (win_store == NULL) {
+        flb_error("[tls] Cannot open cert store: %i", GetLastError());
+        return -1;
+    }
+
+    while (win_cert = CertEnumCertificatesInStore(win_store, win_cert)) {
+        if (win_cert->dwCertEncodingType & X509_ASN_ENCODING) {
+            /*
+             * Decode the certificate into X509 struct.
+             *
+             * The additional pointer variable is necessary per OpenSSL docs because the
+             * pointer is incremented by d2i_X509.
+             */
+            win_cert_data = win_cert->pbCertEncoded;
+            ossl_cert = d2i_X509(NULL, &win_cert_data, win_cert->cbCertEncoded);
+            if (!ossl_cert) {
+                flb_debug("[tls] Cannot parse a certificate. skipping...");
+                continue;
+            }
+
+            /* Add X509 struct to the openssl cert store */
+            ret = X509_STORE_add_cert(ossl_store, ossl_cert);
+            if (!ret) {
+                flb_warn("[tls] Failed to add a certificate to the store: %lu: %s",
+                         ERR_get_error(), ERR_error_string(ERR_get_error(), NULL));
+            }
+            X509_free(ossl_cert);
+        }
+    }
+
+    if (!CertCloseStore(win_store, 0)) {
+        flb_error("[tls] Cannot close cert store: %i", GetLastError());
+        return -1;
+    }
+    return 0;
+}
+#endif
+
 static int load_system_certificates(struct tls_context *ctx)
 {
     int ret;
-    const char ca_path[] = "/etc/ssl/certs/";
+    const char *ca_file = FLB_DEFAULT_SEARCH_CA_BUNDLE;
 
     /* For Windows use specific API to read the certs store */
 #ifdef _MSC_VER
-    //return windows_load_system_certificates(ctx);
+    return windows_load_system_certificates(ctx);
 #endif
+    if (access(ca_file, R_OK) != 0) {
+        ca_file = NULL;
+    }
 
-    if (access(RHEL_DEFAULT_CA, R_OK) == 0) {
-        ret = SSL_CTX_load_verify_locations(ctx->ctx, RHEL_DEFAULT_CA, ca_path);
-    }
-    else {
-        ret = SSL_CTX_load_verify_locations(ctx->ctx, NULL, ca_path);
-    }
+    ret = SSL_CTX_load_verify_locations(ctx->ctx, ca_file, FLB_DEFAULT_CA_DIR);
 
     if (ret != 1) {
         ERR_print_errors_fp(stderr);
@@ -157,15 +197,20 @@ static int load_system_certificates(struct tls_context *ctx)
     return 0;
 }
 
-static void *tls_context_create(int verify, int debug,
+static void *tls_context_create(int verify,
+                                int debug,
+                                int mode,
                                 const char *vhost,
                                 const char *ca_path,
-                                const char *ca_file, const char *crt_file,
-                                const char *key_file, const char *key_passwd)
+                                const char *ca_file,
+                                const char *crt_file,
+                                const char *key_file,
+                                const char *key_passwd)
 {
     int ret;
     SSL_CTX *ssl_ctx;
     struct tls_context *ctx;
+    char err_buf[256];
 
     /*
      * Init library ? based in the documentation on OpenSSL >= 1.1.0 is not longer
@@ -182,9 +227,20 @@ static void *tls_context_create(int verify, int debug,
      *
      * https://www.openssl.org/docs/man1.0.2/man3/SSLv23_method.html
      */
-    ssl_ctx = SSL_CTX_new(SSLv23_client_method());
+    if (mode == FLB_TLS_SERVER_MODE) {
+        ssl_ctx = SSL_CTX_new(SSLv23_server_method());
+    }
+    else {
+        ssl_ctx = SSL_CTX_new(SSLv23_client_method());
+    }
+
 #else
-    ssl_ctx = SSL_CTX_new(TLS_client_method());
+    if (mode == FLB_TLS_SERVER_MODE) {
+        ssl_ctx = SSL_CTX_new(TLS_server_method());
+    }
+    else {
+        ssl_ctx = SSL_CTX_new(TLS_client_method());
+    }
 #endif
     if (!ssl_ctx) {
         flb_error("[openssl] could not create context");
@@ -197,6 +253,7 @@ static void *tls_context_create(int verify, int debug,
         return NULL;
     }
     ctx->ctx = ssl_ctx;
+    ctx->mode = mode;
     ctx->debug_level = debug;
     pthread_mutex_init(&ctx->mutex, NULL);
 
@@ -212,20 +269,18 @@ static void *tls_context_create(int verify, int debug,
     if (ca_path) {
         ret = SSL_CTX_load_verify_locations(ctx->ctx, NULL, ca_path);
         if (ret != 1) {
-            flb_error("[tls] ca_path'%s' %lu: %s",
-                      ca_path,
-                      ERR_get_error(),
-                      ERR_error_string(ERR_get_error(), NULL));
+            ERR_error_string_n(ERR_get_error(), err_buf, sizeof(err_buf)-1);
+            flb_error("[tls] ca_path '%s' %lu: %s",
+                      ca_path, ERR_get_error(), err_buf);
             goto error;
         }
     }
     else if (ca_file) {
         ret = SSL_CTX_load_verify_locations(ctx->ctx, ca_file, NULL);
         if (ret != 1) {
+            ERR_error_string_n(ERR_get_error(), err_buf, sizeof(err_buf)-1);
             flb_error("[tls] ca_file '%s' %lu: %s",
-                      ca_file,
-                      ERR_get_error(),
-                      ERR_error_string(ERR_get_error(), NULL));
+                      ca_file, ERR_get_error(), err_buf);
             goto error;
         }
     }
@@ -236,11 +291,10 @@ static void *tls_context_create(int verify, int debug,
     /* crt_file */
     if (crt_file) {
         ret = SSL_CTX_use_certificate_chain_file(ssl_ctx, crt_file);
-		if (ret != 1) {
+        if (ret != 1) {
+            ERR_error_string_n(ERR_get_error(), err_buf, sizeof(err_buf)-1);
             flb_error("[tls] crt_file '%s' %lu: %s",
-                      crt_file,
-                      ERR_get_error(),
-                      ERR_error_string(ERR_get_error(), NULL));
+                      crt_file, ERR_get_error(), err_buf);
             goto error;
         }
     }
@@ -254,10 +308,9 @@ static void *tls_context_create(int verify, int debug,
         ret = SSL_CTX_use_PrivateKey_file(ssl_ctx, key_file,
                                           SSL_FILETYPE_PEM);
         if (ret != 1) {
+            ERR_error_string_n(ERR_get_error(), err_buf, sizeof(err_buf)-1);
             flb_error("[tls] key_file '%s' %lu: %s",
-                      key_file,
-                      ERR_get_error(),
-                      ERR_error_string(ERR_get_error(), NULL));
+                      crt_file, ERR_get_error(), err_buf);
         }
 
         /* Make sure the key and certificate file match */
@@ -276,7 +329,7 @@ static void *tls_context_create(int verify, int debug,
 }
 
 static void *tls_session_create(struct flb_tls *tls,
-                                struct flb_upstream_conn *u_conn)
+                                int fd)
 {
     struct tls_session *session;
     struct tls_context *ctx = tls->ctx;
@@ -298,9 +351,11 @@ static void *tls_session_create(struct flb_tls *tls,
         pthread_mutex_unlock(&ctx->mutex);
         return NULL;
     }
+
+    session->continuation_flag = FLB_FALSE;
     session->ssl = ssl;
-    session->fd = u_conn->fd;
-    SSL_set_fd(ssl, u_conn->fd);
+    session->fd = fd;
+    SSL_set_fd(ssl, fd);
 
     /*
      * TLS Debug Levels:
@@ -314,7 +369,6 @@ static void *tls_session_create(struct flb_tls *tls,
     if (tls->debug == 1) {
         SSL_set_info_callback(session->ssl, tls_info_callback);
     }
-    SSL_set_connect_state(ssl);
     pthread_mutex_unlock(&ctx->mutex);
     return session;
 }
@@ -333,7 +387,6 @@ static int tls_session_destroy(void *session)
 
     if (flb_socket_error(ptr->fd) == 0) {
         SSL_shutdown(ptr->ssl);
-        SSL_shutdown(ptr->ssl);
     }
     SSL_free(ptr->ssl);
     flb_free(ptr);
@@ -343,26 +396,57 @@ static int tls_session_destroy(void *session)
     return 0;
 }
 
-static int tls_net_read(struct flb_upstream_conn *u_conn,
+static int tls_net_read(struct flb_tls_session *session,
                         void *buf, size_t len)
 {
     int ret;
-    struct tls_session *session = (struct tls_session *) u_conn->tls_session;
+    char err_buf[256];
     struct tls_context *ctx;
+    struct tls_session *backend_session;
+
+    if (session->ptr == NULL) {
+        flb_error("[tls] error: uninitialized backend session");
+
+        return -1;
+    }
+
+    backend_session = (struct tls_session *) session->ptr;
+
+    ctx = backend_session->parent;
 
-    ctx = session->parent;
     pthread_mutex_lock(&ctx->mutex);
 
     ERR_clear_error();
-    ret = SSL_read(session->ssl, buf, len);
+
+    ret = SSL_read(backend_session->ssl, buf, len);
+
     if (ret <= 0) {
-        ret = SSL_get_error(session->ssl, ret);
+        ret = SSL_get_error(backend_session->ssl, ret);
+
         if (ret == SSL_ERROR_WANT_READ) {
             ret = FLB_TLS_WANT_READ;
         }
         else if (ret == SSL_ERROR_WANT_WRITE) {
             ret = FLB_TLS_WANT_WRITE;
         }
+        else if (ret == SSL_ERROR_SYSCALL) {
+            flb_errno();
+            ERR_error_string_n(ret, err_buf, sizeof(err_buf)-1);
+            flb_error("[tls] syscall error: %s", err_buf);
+
+            /* According to the documentation these are non-recoverable
+             * errors so we don't need to screen them before saving them
+             * to the net_error field.
+             */
+
+            session->connection->net_error = errno;
+
+            ret = -1;
+        }
+        else if (ret < 0) {
+            ERR_error_string_n(ret, err_buf, sizeof(err_buf)-1);
+            flb_error("[tls] error: %s", err_buf);
+        }
         else {
             ret = -1;
         }
@@ -372,30 +456,59 @@ static int tls_net_read(struct flb_upstream_conn *u_conn,
     return ret;
 }
 
-static int tls_net_write(struct flb_upstream_conn *u_conn,
+static int tls_net_write(struct flb_tls_session *session,
                          const void *data, size_t len)
 {
     int ret;
+    char err_buf[256];
     size_t total = 0;
-    struct tls_session *session = (struct tls_session *) u_conn->tls_session;
     struct tls_context *ctx;
+    struct tls_session *backend_session;
+
+    if (session->ptr == NULL) {
+        flb_error("[tls] error: uninitialized backend session");
+
+        return -1;
+    }
+
+    backend_session = (struct tls_session *) session->ptr;
+    ctx = backend_session->parent;
 
-    ctx = session->parent;
     pthread_mutex_lock(&ctx->mutex);
 
     ERR_clear_error();
-    ret = SSL_write(session->ssl,
+
+    ret = SSL_write(backend_session->ssl,
                     (unsigned char *) data + total,
                     len - total);
+
     if (ret <= 0) {
-        ret = SSL_get_error(session->ssl, ret);
+        ret = SSL_get_error(backend_session->ssl, ret);
+
         if (ret == SSL_ERROR_WANT_WRITE) {
             ret = FLB_TLS_WANT_WRITE;
         }
         else if (ret == SSL_ERROR_WANT_READ) {
             ret = FLB_TLS_WANT_READ;
         }
+        else if (ret == SSL_ERROR_SYSCALL) {
+            flb_errno();
+            ERR_error_string_n(ret, err_buf, sizeof(err_buf)-1);
+            flb_error("[tls] syscall error: %s", err_buf);
+
+            /* According to the documentation these are non-recoverable
+             * errors so we don't need to screen them before saving them
+             * to the net_error field.
+             */
+
+            session->connection->net_error = errno;
+
+            ret = -1;
+        }
         else {
+            ERR_error_string_n(ret, err_buf, sizeof(err_buf)-1);
+            flb_error("[tls] error: %s", err_buf);
+
             ret = -1;
         }
     }
@@ -406,40 +519,85 @@ static int tls_net_write(struct flb_upstream_conn *u_conn,
     return ret;
 }
 
-static int tls_net_handshake(struct flb_tls *tls, void *ptr_session)
+static int tls_net_handshake(struct flb_tls *tls,
+                             char *vhost,
+                             void *ptr_session)
 {
     int ret = 0;
+    char err_buf[256];
     struct tls_session *session = ptr_session;
     struct tls_context *ctx;
 
     ctx = session->parent;
     pthread_mutex_lock(&ctx->mutex);
 
-    if (tls->vhost) {
-        SSL_set_tlsext_host_name(session->ssl, tls->vhost);
+    if (!session->continuation_flag) {
+        if (tls->mode == FLB_TLS_CLIENT_MODE) {
+            SSL_set_connect_state(session->ssl);
+        }
+        else if (tls->mode == FLB_TLS_SERVER_MODE) {
+            SSL_set_accept_state(session->ssl);
+        }
+        else {
+            flb_error("[tls] error: invalid tls mode : %d", tls->mode);
+            pthread_mutex_unlock(&ctx->mutex);
+            return -1;
+        }
+
+        if (vhost != NULL) {
+            SSL_set_tlsext_host_name(session->ssl, vhost);
+        }
+        else if (tls->vhost) {
+            SSL_set_tlsext_host_name(session->ssl, tls->vhost);
+        }
     }
 
     ERR_clear_error();
-    ret = SSL_connect(session->ssl);
+
+    if (tls->mode == FLB_TLS_CLIENT_MODE) {
+        ret = SSL_connect(session->ssl);
+    }
+    else if (tls->mode == FLB_TLS_SERVER_MODE) {
+        ret = SSL_accept(session->ssl);
+    }
+
     if (ret != 1) {
         ret = SSL_get_error(session->ssl, ret);
         if (ret != SSL_ERROR_WANT_READ &&
             ret != SSL_ERROR_WANT_WRITE) {
             ret = SSL_get_error(session->ssl, ret);
+            // The SSL_ERROR_SYSCALL with errno value of 0 indicates unexpected
+            //  EOF from the peer. This is fixed in OpenSSL 3.0.
+            if (ret == 0) {
+            	flb_error("[tls] error: unexpected EOF");
+            } else {
+                ERR_error_string_n(ret, err_buf, sizeof(err_buf)-1);
+                flb_error("[tls] error: %s", err_buf);
+            }
+
             pthread_mutex_unlock(&ctx->mutex);
+
             return -1;
         }
 
         if (ret == SSL_ERROR_WANT_WRITE) {
             pthread_mutex_unlock(&ctx->mutex);
+
+            session->continuation_flag = FLB_TRUE;
+
             return FLB_TLS_WANT_WRITE;
         }
         else if (ret == SSL_ERROR_WANT_READ) {
             pthread_mutex_unlock(&ctx->mutex);
+
+            session->continuation_flag = FLB_TRUE;
+
             return FLB_TLS_WANT_READ;
         }
     }
 
+    session->continuation_flag = FLB_FALSE;
+
     pthread_mutex_unlock(&ctx->mutex);
     flb_trace("[tls] connection and handshake OK");
     return 0;
@@ -447,12 +605,12 @@ static int tls_net_handshake(struct flb_tls *tls, void *ptr_session)
 
 /* OpenSSL backend registration */
 static struct flb_tls_backend tls_openssl = {
-    .name            = "openssl",
-    .context_create  = tls_context_create,
-    .context_destroy = tls_context_destroy,
-    .session_create  = tls_session_create,
-    .session_destroy = tls_session_destroy,
-    .net_read        = tls_net_read,
-    .net_write       = tls_net_write,
-    .net_handshake   = tls_net_handshake
+    .name                 = "openssl",
+    .context_create       = tls_context_create,
+    .context_destroy      = tls_context_destroy,
+    .session_create       = tls_session_create,
+    .session_destroy      = tls_session_destroy,
+    .net_read             = tls_net_read,
+    .net_write            = tls_net_write,
+    .net_handshake        = tls_net_handshake,
 };
